/***********************************************************
        Copyright 1996 by Carnegie Mellon University

                      All Rights Reserved

Permission to use, copy, modify, and distribute this software and its
documentation for any purpose and without fee is hereby granted,
provided that the above copyright notice appear in all copies and that
both that copyright notice and this permission notice appear in
supporting documentation, and that the name of CMU not be
used in advertising or publicity pertaining to distribution of the
software without specific, written prior permission.

CMU DISCLAIMS ALL WARRANTIES WITH REGARD TO THIS SOFTWARE, INCLUDING
ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS, IN NO EVENT SHALL
CMU BE LIABLE FOR ANY SPECIAL, INDIRECT OR CONSEQUENTIAL DAMAGES OR
ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS,
WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION,
ARISING OUT OF OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS
SOFTWARE.
******************************************************************/

#include "defs.h"

#include <stdio.h>
#include <string.h>
#include <ctype.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <sys/time.h>
#include <netinet/in.h>
#include <net/if.h>
#include <net/route.h>
#include <netdb.h>
#include <errno.h>
#include <sys/stat.h>	/* for stat, open */
#include <stdio.h> /* for fdopen */

#ifdef HAVE_UNISTD_H
#include <unistd.h> /* for setuid/seteuid/setgid/setegid, unlink */
#endif

#ifdef HAVE_FCNTL_H
#include <fcntl.h> /* for open */
#else
#ifdef HAVE_SYS_FILE_H
#include <sys/file.h> /* for open, O_ constants defined here on BSD systems */
#endif
#endif


#include "defaults.h"
#include "bootp.h"
#include "bootpd.h"
#include "dhcpd.h"
#include "dhcp.h"
#include "utils.h"
#include "my_syslog.h"

int insert_byte_opt(bp,opt,tag,value)
struct bootp *bp;
struct opts *opt;
unsigned char tag;
unsigned char value;
{
	int i;
	unsigned char *cp;

	cp=(unsigned char*)bp;
	if(opt->cur==0)
		opt->cur=VEND_OFFSET;
	
	if(opt->cur+7>opt->len){
		cp[opt->cur++]=TAG_DHCP_OVL;
		cp[opt->cur++]=1;
		cp[opt->cur++]=1; /* file overload */
		cp[opt->cur]=TAG_END;
		opt->cur=FILE_OFFSET;
	} else if (opt->cur<SNAME_OFFSET &&opt->cur+7>SNAME_OFFSET) {
		cp[opt->cur]=TAG_END;
		for(i=4;(u_long)bp->bp_vend[i]!=TAG_DHCP_OVL;i++);
		bp->bp_vend[i+2]=3;
		opt->cur=SNAME_OFFSET;
	} else if(opt->cur<VEND_OFFSET && opt->cur+7>VEND_OFFSET)
		return -1;
	cp[opt->cur++]=tag;
	cp[opt->cur++]=1;
	cp[opt->cur++]=value;

	return 0;
}

/* Insert 1-byte 'tag' and 4-byte 'value' into the packet represented by 'bp' and 'opt'.
   We try to put it in 'opt' if it fits, else overload it into an unused 'bp' header field.
   Note that 'value' is copied in byte-for-byte; if any network-byte ordering conversion needs
   to be done, it must be done before calling us */
int insert_ulong_opt(bp,opt,tag,value)
struct bootp *bp;
struct opts *opt;
unsigned char tag;
unsigned long value;
{
	int i;
	unsigned char *cp;

	cp=(u_char*)bp;
	if(opt->cur==0)
		opt->cur=VEND_OFFSET;
	
	if(opt->cur+11>opt->len){
		cp[opt->cur++]=TAG_DHCP_OVL;
		cp[opt->cur++]=1;
		cp[opt->cur++]=1; /* file overload */
		cp[opt->cur]=TAG_END;
		opt->cur=FILE_OFFSET;
	} else if (opt->cur+11>SNAME_OFFSET && opt->cur<SNAME_OFFSET){
		cp[opt->cur]=TAG_END;
		for(i=4;(u_long)bp->bp_vend[i]!=TAG_DHCP_OVL;i++);
		bp->bp_vend[i+2]=3;
		opt->cur=SNAME_OFFSET;
	} else if(opt->cur<VEND_OFFSET && opt->cur+11>VEND_OFFSET)
		return -1;
	cp[opt->cur++]=tag;
	cp[opt->cur++]=4;
	bcopy(&value,cp+opt->cur,4);
	opt->cur+=4;

	return 0;
}

/* Insert 1-byte 'tag' and 2-byte 'value' into the packet represented by 'bp' and 'opt'.
   We try to put it in 'opt' if it fits, else overload it into an unused 'bp' header field.
   Note that 'value' is copied in byte-for-byte; if any network-byte ordering conversion needs
   to be done, it must be done before calling us */
int insert_ushort_opt(bp,opt,tag,value)
struct bootp *bp;
struct opts *opt;
unsigned char tag;
unsigned short value;
{
	int i;
	unsigned char *cp;

	cp=(u_char*)bp;
	if(opt->cur==0)
		opt->cur=VEND_OFFSET;
	
	if(opt->cur+9>opt->len){
		cp[opt->cur++]=TAG_DHCP_OVL;
		cp[opt->cur++]=1;
		cp[opt->cur++]=1; /* file overload */
		cp[opt->cur]=TAG_END;
		opt->cur=FILE_OFFSET;
	} else if (opt->cur+9>SNAME_OFFSET && opt->cur<SNAME_OFFSET){
		cp[opt->cur]=TAG_END;
		for(i=4;(u_long)bp->bp_vend[i]!=TAG_DHCP_OVL;i++);
		bp->bp_vend[i+2]=3;
		opt->cur=SNAME_OFFSET;
	} else if(opt->cur<VEND_OFFSET && opt->cur+9>VEND_OFFSET)
		return -1;
	cp[opt->cur++]=tag;
	cp[opt->cur++]=2;
	bcopy(&value,cp+opt->cur,2);
	opt->cur+=2;

	return 0;
}

int insert_addrs_opt(bp,opt,tag,value,len)
struct bootp *bp;
struct opts *opt;
unsigned char tag;
struct in_addr *value;
int len;
{
	int i;
	unsigned char *cp;
	int bytes;

	cp=(u_char*)bp;
	if(opt->cur==0)
		opt->cur=VEND_OFFSET;
	bytes=len*sizeof(struct in_addr);	
	if(opt->cur+6+bytes>opt->len){
		cp[opt->cur++]=TAG_DHCP_OVL;
		cp[opt->cur++]=1;
		cp[opt->cur++]=1; /* file overload */
		cp[opt->cur]=TAG_END;
		opt->cur=FILE_OFFSET;
	} else if (opt->cur+6+bytes>SNAME_OFFSET && opt->cur<SNAME_OFFSET){
		cp[opt->cur]=TAG_END;
		for(i=4;(u_long)bp->bp_vend[i]!=TAG_DHCP_OVL;i++);
		bp->bp_vend[i+2]=3;
		opt->cur=SNAME_OFFSET;
	} else if(opt->cur<VEND_OFFSET && opt->cur+6+bytes>VEND_OFFSET)
		return -1;
	cp[opt->cur++]=tag;
	cp[opt->cur++]=bytes;
	bcopy(value,cp+opt->cur,bytes);
	opt->cur+=bytes;

	return 0;
}

int insert_end_opt(bp,opt)
struct bootp *bp;
struct opts *opt;
{
	unsigned char *cp;
	cp=(u_char*)bp;
	cp[opt->cur++]=TAG_END;
	return 0;
}
int insert_string_opt(bp,opt,tag,value,len)
struct bootp *bp;
struct opts *opt;
unsigned char tag;
char *value;
int len;
{
	int i;
	unsigned char *cp;
	int bytes;

	cp=(u_char*)bp;
	if(opt->cur==0)
		opt->cur=VEND_OFFSET;
	bytes=len;	
	if(opt->cur+6+bytes>opt->len){
		cp[opt->cur++]=TAG_DHCP_OVL;
		cp[opt->cur++]=1;
		cp[opt->cur++]=1; /* file overload */
		cp[opt->cur]=TAG_END;
		opt->cur=FILE_OFFSET;
	} else if (opt->cur+6+bytes>SNAME_OFFSET && opt->cur<SNAME_OFFSET) {
		cp[opt->cur]=TAG_END;
		for(i=4;(u_long)bp->bp_vend[i]!=TAG_DHCP_OVL;i++);
		bp->bp_vend[i+2]=3;
		opt->cur=SNAME_OFFSET;
	} else if(opt->cur<VEND_OFFSET && opt->cur+6+bytes>VEND_OFFSET)
		return -1;
	cp[opt->cur++]=tag;
	cp[opt->cur++]=bytes;
	bcopy(value,cp+opt->cur,bytes);
	opt->cur+=bytes;

	return 0;
}

int insert_str_opt(bp,opt,tag,value)
struct bootp *bp;
struct opts *opt;
unsigned char tag;
char *value;
{
	return insert_string_opt(bp,opt,tag,value,strlen(value));
}

/* smalloc()  --  safe malloc()
 *
 * Always returns a valid pointer (if it returns at all).  The allocated
 * memory is initialized to all zeros.  If malloc() returns an error, a
 * message is printed using the report() function and the program aborts
 * with a status of 1.
 */

void *
smalloc(nbytes)
	unsigned nbytes;
{
	void *retvalue;

	retvalue = malloc(nbytes);
	if (!retvalue) {
		report(LOG_ERR, "malloc() failure -- exiting");
		exit(1);
	}
	bzero((char *) retvalue, nbytes);
	return retvalue;
}


/* change_eid() -- change effective uid or gid
 *
 * Change effective uid or gid to id specified.
 * Return 0 on success, non-zero on error.
 * Only supported if HAVE_SETEUID available, to ensure we can change back later.
 */
int change_eid(uid_or_gid, uid, gid)
int uid_or_gid;		/* 1 == uid, 2==gid */
uid_t uid;			/* if uid_or_gid==1, this is uid to change to; else this is ignored */
gid_t gid;			/* if uid_or_gid==2, this is gid to change to; else this is ignored */
{
	int rc;
	long tmp_id; /* for error reporting */
	static char tmp_str[256];	/* for error reporting */
	char * funcname_ptr; /* for error reporting, will point to one of the following: */
	static char * funcname_seteuid="seteuid";
	static char * funcname_setegid="setegid";

	if ((uid_or_gid != 1) && (uid_or_gid != 2)) {
		report(LOG_ERR, "change_id(): uid_or_gid flag has invalid value: %d", uid_or_gid);
		return -1;
	}

#ifdef HAVE_SETEUID
	/* sete*id() changes only the e*id, leaving real *id (and saved set *id, if it exists) unchanged.
	   Since this server must be started by root, the real *id remains 0, so we'll always be able to
	   change back and forth. */
	if (uid_or_gid==1) {
		funcname_ptr = funcname_seteuid;
		tmp_id = uid;
		rc=seteuid(uid);
	} else {
		funcname_ptr = funcname_setegid;
		tmp_id = gid;
		rc=setegid(gid);
	}
#else
		report(LOG_ERR, "change_id(): HAVE_SETEUID not defined, huh?");
		return -1;
#endif /* HAVE_SETEUID */

	if (rc) {
		sprintf(tmp_str, "%s(%d) failed: %s", funcname_ptr, tmp_id, get_errmsg());
		report(LOG_ERR, tmp_str);
		return -1;
	}

	return 0;
}

/* open_for_writing() - open a regular file for writing.
 * If the file already exists, it will be removed first.
 * On error we return NULL and log an error message.
 * Based closely on write_open() from BIND 8.1.2.
 */
FILE *
open_for_writing(filename) 
char * filename;
{
	int fd;
	FILE *stream;
	struct stat stat_buffer;

	if (stat(filename, &stat_buffer) < 0) {
		if (errno != ENOENT) {
			report(LOG_ERR, "open_for_writing(): stat of %s failed: %s", filename, get_errmsg());
			return NULL;
		} 
	} else { /* filename already exists */
		if (!(stat_buffer.st_mode & S_IFREG)) {
			/* since we just want to open a regular file for writing, its existance as a special
			   file probably means something is wrong, so let's TRY to avoid removing it. */
			report(LOG_ERR, "open_for_writing(): %s exists but isn't a regular file", filename);
			return(NULL);
		}
	}

	/* At this point, filename either does not exist, or already exists as a regular file.
	   Either way, it's toast.  Note that it is possible for it's status to have changed since
	   we checked earlier; e.g. it may have changed from non-existance to existance as a special file.
	   So it's still possible we'll wipe out a special file, but we've made an effort to avoid it. */
	if (unlink(filename) < 0) {
		if (errno != ENOENT) {
			report(LOG_ERR, "open_for_writing(): unlink of %s failed: %s", filename, get_errmsg());
			return NULL;
		}
	}

	/* Use open() to get the atomic behavior provided by O_CREAT|O_EXCL.  If it succeeds,
	   convert the file descriptor into a file stream, which is what we're supposed to return. */
	fd = open(filename, O_WRONLY|O_CREAT|O_EXCL, S_IRUSR|S_IWUSR|S_IRGRP|S_IROTH);
	if (fd < 0) {
		report(LOG_ERR, "open_for_writing(): open of %s failed: %s", filename, get_errmsg());
		return NULL;
	}
	stream = fdopen(fd, "w");
	if (stream == NULL) {
		report(LOG_ERR, "open_for_writing(): fdopen failed: %s", get_errmsg());
		(void) close(fd);
		return NULL;
	}
	return(stream);
}
